import numpy as np
import scipy as sp
from scipy import optimize
from scipy import interpolate
from scipy import integrate
import functools
import time
import copy

import os
from scipy import fft

from thomasfermi.algo import TFdata
import thomasfermi.FQHE as FQHE

class FQHE_TFdata(TFdata):

    def __init__(self, pars):
        """
            Base class for representing the Thomas-Fermi calculation.
            Input dictionary pars encodes all the relevant parameters.

            l_B - magnetic length
            L - system size in l_B
            dx, dy - mesh size in l_B

            d_t, d_b - distance (in nm) from 2DEG to top/bottom gates

            eps_xy, eps_z - anisotropic permitivities depending on the material
                filling the space between the 2DEG and gates.

            Ec - strength of the Coulomb term in eV

            a - lattice spacing of graphene

        """

        self.dx = pars.setdefault('dx', 0.5)
        self.dy = pars.setdefault('dy', 0.5)
        self.l_B = pars.setdefault('l_B', 9.2)

        self.energy_scale = pars.setdefault('energy_scale', 100)
        self.n_lower = pars.setdefault('n_lower', -0.08)
        self.n_upper = pars.setdefault('n_upper', 1.08)

        super().__init__(pars)

        self.Sq = self.gaussian_conv()/(self.dx*self.dy/self.a**2)

        self.mu_factor = pars.setdefault('mu_factor', 1)

        self.mu_model = pars.setdefault('mu_model', 'expt')

        if self.mu_model == 'expt':
            self.make_mu_experimental(self.mu_factor)
        elif self.mu_model == 'phenom':
            self.make_mu_phenom()
        else:
            raise ValueError("mu model does not exist")

    def gaussian_conv(self):
        """
            Helper function for constructing the Gaussian convolution.
        """

        Q = (self.qx,self.qy)

        Q = np.linalg.norm(Q, axis = 0)

        ans = np.exp(-(Q/self.a)**2/2.)

        return ans/(np.sum(ans)*self.dqx*self.dqy)

    def make_mu_experimental(self, mu_factor):
        """
            Load experimental data from arXiv: 2008.05466
            Create interpolated functions for dmu/dn, mu, and E_int
        """

        d = np.loadtxt(os.path.dirname(FQHE.__file__) + "\\nu_to_mu.dat")
        d[:,1] *= self.mu_factor*1.
        self.mu = interpolate.interp1d(d[:,0], d[:,1])
        self.dmu = interpolate.interp1d(0.5 * (d[1:,0] + d[:-1,0]), np.diff(self.mu(d[:,0])))
        integrated_mu = integrate.cumtrapz(self.mu(d[:,0]), x=d[:,0], initial=0.)
        integrated_mu -= integrated_mu[np.argmin(np.abs(d[:,0]))]
        self.E_int = interpolate.interp1d(d[:,0], integrated_mu)

        self.T = np.abs(self.E_int(1) - self.E_int(0))

        if self.n_lower == None:
            self.n_lower = -1.
            self.n_upper = 1.9

        return True

    def make_mu_phenom(self, dos_gap = 0.6, dos_plateau = 1.1, dos_frac_plateau=0.15, dn=1e-4):
        """
            Construct a phenomenological model for mu involving stepwise
            constant compressibility.
        """

        dn_plateau = 0.05
        dn_frac_plateau = 0.025
        n = np.arange(-2,2+1e-16,dn)

        plateau = np.abs(n - np.round(n)) < dn_plateau
        frac_plateau = (np.abs((n - 1./3) % 1) < dn_frac_plateau) + (np.abs((n - 2./3) % 1) < dn_frac_plateau)
        dos = n*0 - dos_gap
        dos[plateau] = dn_plateau / dos_plateau
        dos[frac_plateau] = dn_frac_plateau / dos_frac_plateau
        dmu_vec = 1. / dos

        mu = sp.integrate.cumtrapz(dmu_vec, dx=dn, initial=0.)
        mu -= mu[np.argmin(np.abs(n))]
        mu = sp.interpolate.interp1d(n, mu)
        dmu = sp.interpolate.interp1d(n, dmu_vec)
        E_int = sp.integrate.cumtrapz(mu(n), dx=dn, initial=0.)
        E_int -= E_int[np.argmin(np.abs(n))]
        E_int = sp.interpolate.interp1d(n, E_int)
        self.dmu = dmu
        self.mu = mu
        self.E_int = E_int

    




import numpy as np
from scipy import fft

class make_Phi(object):

    def __init__(self, pars):
        """
            Parent class to construct the potential Phi_ext felt by the 2DEG due to the gates.
            Input is a TFdata that contains the calculation setup and device geometry.
        """

        self.TFdata = pars["TFdata"]

        self.L = self.TFdata.L

        self.Nx = self.TFdata.Nx
        self.Ny = self.TFdata.Ny

        if self.TFdata.l_B != None:
            self.l_B = self.TFdata.l_B
        else:
            self.l_B = 1

        self.X = self.TFdata.X
        self.Y = self.TFdata.Y

        self.d_t = self.TFdata.d_t
        self.d_b = self.TFdata.d_b

        self.Gq_g = self.TFdata.Gq_g

    def gate_to_ext(self, Phi_gate):
        """
            Converts the gate potential to the external potential.
        """

        if self.Phi_back == None:
            self.Phi_back = 0

        self.Phi_ext = self.Phi_back/(self.d_t+self.d_b)*self.d_t + np.real((fft.ifft2(fft.fftshift(self.Gq_g)*fft.fft2(self.Phi_gate))).reshape((-1,)))

        return True


class make_QPC_Phi(make_Phi):

    def __init__(self, pars):
        """
            Construct the potential Phi_ext felt by the 2DEG due to QPC gates.
            Input is a TFdata and additional parameters listed below:

            w_EW, w_NS - displacement from the center of the E/W and N/S gates.
            dn_NS, dn_EW - detuning of the potential on the corresponding pair of gates
            Phi_EW, Phi_NS - potential at which the corresponding gates are held
            Phi_back - potential at which the back gate is held
        """

        super().__init__(pars)

        self.w_EW = pars.setdefault('w_EW', 35)
        self.w_NS = pars.setdefault('w_NS', 35)
        self.w_EW = self.w_EW/self.l_B
        self.w_NS = self.w_NS/self.l_B

        self.dn_NS = pars.setdefault('dn_NS', 0)
        self.dn_EW = pars.setdefault('dn_EW', 0)

        self.Phi_EW = pars.setdefault('Phi_EW', -0.8)
        self.Phi_NS = pars.setdefault('Phi_NS', 0.2)
        self.Phi_back = pars.setdefault('Phi_back', -0.2)

        self.Phi_EW_ext = pars.setdefault('Phi_EW_ext', None)
        self.Phi_NS_ext = pars.setdefault('Phi_NS_ext', None)

        if self.Phi_EW_ext != None:
            self.Phi_EW = (self.Phi_EW_ext - self.d_t/(self.d_t + self.d_b)*self.Phi_back)/(self.d_b/(self.d_t + self.d_b))
            self.Phi_NS = (self.Phi_NS_ext - self.d_t/(self.d_t + self.d_b)*self.Phi_back)/(self.d_b/(self.d_t + self.d_b))

        self.Phi_gate = self.Phi_EW*(1+self.dn_EW)*(self.X < -self.w_EW)*(self.Y < -self.w_EW)*(self.X > -self.L + self.w_EW)*(self.Y > -self.L + self.w_EW) \
                + self.Phi_NS*(1+self.dn_NS)*(self.X > self.w_NS)*(self.Y < -self.w_NS)*(self.X < self.L - self.w_NS)*(self.Y > -self.L + self.w_NS) \
                + self.Phi_NS*(1-self.dn_NS)*(self.X < -self.w_NS)*(self.Y > self.w_NS)*(self.X > -self.L + self.w_NS)*(self.Y < self.L - self.w_NS) \
                + self.Phi_EW*(1-self.dn_EW)*(self.X > self.w_EW)*(self.Y > self.w_EW)*(self.X < self.L - self.w_EW)*(self.Y < self.L - self.w_EW)

        self.gate_to_ext(self.Phi_gate)

class make_uniform_Phi(make_Phi):

    def __init__(self, pars):
        """
            Construct the potential Phi_ext felt by the 2DEG due to a uniform gate.
            Input is a TFdata and additional parameters listed below:

            Phi_top - potential at which the top gate is held
            Phi_back - potential at which the back gate is held
        """

        super().__init__(pars)

        self.Phi_back = pars.setdefault('Phi_back', -0.2)
        self.Phi_top = pars.setdefault('Phi_top', 0)

        self.Phi_ext = self.Phi_back/(self.d_t+self.d_b)*self.d_t + self.Phi_top/(self.d_t+self.d_b)*self.d_b + 0*(self.X).reshape((-1,))

class make_gradient_Phi(make_Phi):

    def __init__(self, pars):
        """
            Construct the potential Phi_ext felt by the 2DEG due to a uniform gate.
            Input is a TFdata and additional parameters listed below:

            Phi_top - potential at which the top gate is held
            Phi_back - potential at which the back gate is held
        """
        super().__init__(pars)

        self.Phi_low = pars.setdefault('Phi_low', -0.8)
        self.Phi_high = pars.setdefault('Phi_high', 0.2)
        self.Phi_back = pars.setdefault('Phi_back', -0.2)

        self.Phi_low_ext = pars.setdefault('Phi_low_ext', None)
        self.Phi_high_ext = pars.setdefault('Phi_high_ext', None)

        if self.Phi_low_ext != None:
            self.Phi_low = (self.Phi_low_ext - self.d_t/(self.d_t + self.d_b)*self.Phi_back)/(self.d_b/(self.d_t + self.d_b))
            self.Phi_high = (self.Phi_high_ext - self.d_t/(self.d_t + self.d_b)*self.Phi_back)/(self.d_b/(self.d_t + self.d_b))

        self.Phi_ave = (self.Phi_low + self.Phi_high)/2
        self.Phi_diff = (self.Phi_high - self.Phi_low)/2
        
        self.Phi_gate = self.Phi_ave + self.Phi_diff*np.arange(-1,1,2/(self.Nx//2))*np.ones((self.Nx//2,1))

        self.Phi_gate = self.symmetrize_Phi(self.Phi_gate)

        self.gate_to_ext(self.Phi_gate)

    def symmetrize_Phi(self, Phi):

        Phi = np.concatenate((Phi,Phi[::-1,:]), axis=0)
        Phi = np.concatenate((Phi,Phi[:,::-1]), axis=1)

        Phi = np.roll(np.roll(Phi, self.Nx//4, axis = 0), self.Ny//4, axis = 1)

        return Phi


import numpy as np
import scipy as sp
from scipy import optimize
from scipy import interpolate
from scipy import integrate
import functools
import time
import copy

from scipy import fft

class TFdata(object):

    def __init__(self, pars):
        """
            Base class for representing the Thomas-Fermi calculation.
            Input dictionary pars encodes all the relevant parameters.

            L - system size
            dx, dy - mesh size

            l_B - default length scale in nm, only for the FQHE model

            d_t, d_b - distance (in nm) from 2DEG to top/bottom gates

            eps_xy, eps_z - anisotropic permitivities depending on the material
                filling the space between the 2DEG and gates.

            Ec - strength of the Coulomb term in eV
            a - lattice spacing

        """
              
        self.L = pars.setdefault('L', 20)
        self.dx = pars.setdefault('dx', 1) # don't change unless for the FQHE model
        self.dy = pars.setdefault('dy', 1) # don't change unless for the FQHE model

        self.l_B = pars.setdefault('l_B', 0.246) # default length scale in nm, only for the FQHE model

        self.x = np.arange(-self.L, self.L-1e-8, self.dx) + self.dx/2
        self.y = np.arange(-self.L, self.L-1e-8, self.dy) + self.dy/2
        self.Nx = len(self.x)
        self.Ny = len(self.y)
        self.X, self.Y = np.meshgrid(self.x, self.y)
        self.R2 = self.X**2 + self.Y**2

        if (self.Nx//4)*4 != self.Nx:
            raise ValueError('Wrong Nx, Ny.')

        self.d_t = pars.setdefault('d_t', 65)
        self.d_b = pars.setdefault('d_b', 35)
        self.d_t = self.d_t/self.l_B
        self.d_b = self.d_b/self.l_B


        # Make G

        self.eps = pars.setdefault('eps', 6.85)

        # Default values for hBN
        self.eps_z = pars.setdefault('eps_z', 3.)
        self.eps_xy = pars.setdefault('eps_xy', 6.6)

        # Default value for graphene
        self.Ec = pars.setdefault('Ec', 5.85)

        self.energy_scale = pars.setdefault('energy_scale', 1)

        # Default value for graphene
        self.a = pars.setdefault('a', 0.246)
        self.a = self.a/self.l_B

        self.dqx = np.pi*self.a/self.L
        self.dqy = np.pi*self.a/self.L
        self.qx, self.qy = np.meshgrid(np.arange(-np.pi*self.a/self.dx, np.pi*self.a/self.dx-1e-8, self.dqx), np.arange(-np.pi*self.a/self.dy, np.pi*self.a/self.dy-1e-8, self.dqy))

        self.Gq = self.Ec*self.screened_coulomb()/(2*self.L/self.a)**2

        self.Gq_g = self.screened_coulomb_gate()

        # Make convolution mask, default off, only for FQHE model

        self.S0 = np.exp(-self.R2/2)/(np.sum(np.exp(-self.R2))*self.dx*self.dy)

        self.Sq = (1 + 0*self.qx)/(self.dx*self.dy/self.a**2)

        self.Sq = self.Sq/(np.sum(self.Sq)*self.dqx*self.dqy)

        self.n_lower = pars.setdefault('n_lower', None)
        self.n_upper = pars.setdefault('n_upper', None)

    def logsinh(self, x, oflow = 1e25):

            xflat = x.reshape((-1,))
            mask = xflat<np.log(oflow)

            ans = xflat - np.log(2)
            ans[mask] = np.log(np.sinh(xflat[mask]))

            return ans.reshape((self.Nx,self.Ny))

    def screened_coulomb(self):
        """
            Construct the screened Coulomb matrix in momentum space for self.Gq
        """

        Q = (self.qx,self.qy)
        D = (self.d_t + self.d_b)/2/self.a

        epsf = np.sqrt(self.eps_xy / self.eps_z)

        smear = 10 # in a
        Q = np.linalg.norm(Q, axis = 0)
        # ans = 4*np.pi*np.sinh(self.d_t/self.a*Q*epsf)*np.sinh(self.d_b/self.a*Q*epsf)/np.sinh(2*D*Q*epsf)/Q

        s1 = self.logsinh(self.d_t/self.a*Q*epsf)
        s2 = self.logsinh(self.d_b/self.a*Q*epsf)
        s3 = self.logsinh(2*D*Q*epsf)

        ans = 4*np.pi*np.exp(s1 + s2 - s3) / Q

        ans *= np.exp(-(Q*smear)**2/2.)
        ans[Q == 0] = 2*np.pi*D

        ans /= np.sqrt(self.eps_xy * self.eps_z)

        return ans

    def screened_coulomb_gate(self):
        """
            Construct the screened Coulomb potential from the gate in momentum space.
        """

        Q = (self.qx,self.qy)
        D = (self.d_t + self.d_b)/2/self.a

        epsf = np.sqrt(self.eps_xy / self.eps_z)

        Q = np.linalg.norm(Q, axis = 0)
        s2 = self.logsinh(self.d_b/self.a*Q*epsf)
        s3 = self.logsinh(2*D*Q*epsf)
        ans = np.exp(s2 - s3)

        return ans


class TF_engine(object):

    def __init__(self, pars):
        """
            Class for running the optimization.

            Input is a TFdata with desired parameters already set.

            New parameters are:
                n_iter - number of iterations to run for global optimizer (basinhopping)

        """

        self.TFdata = pars["TFdata"]

        self.L = self.TFdata.L

        self.Nx = self.TFdata.Nx
        self.Ny = self.TFdata.Ny

        self.dx = self.TFdata.dx
        self.dy = self.TFdata.dy

        self.X = self.TFdata.X
        self.Y = self.TFdata.Y

        self.E_int = self.TFdata.E_int
        self.Gq = self.TFdata.Gq
        self.mu = self.TFdata.mu
        self.T = self.TFdata.T

        self.Sq = self.TFdata.Sq
        self.Sq = self.Sq/self.make_n(self.init_n(eta = 0, uniform = True))[0]

        self.Phi_ext = pars['Phi_ext']

        self.energy_scale = self.TFdata.energy_scale

        self.n_lower = self.TFdata.n_lower
        self.n_upper = self.TFdata.n_upper

        self.n_iter = pars.setdefault('n_iter', 20)


    def make_n(self, n0):
        """
            Construct the (convolved) density n on the enlarged system from n
            given on the reduced system.
        """

        n = copy.deepcopy(n0)

        n = n.reshape((self.Nx//2, self.Ny//2))

        n = np.concatenate((n,n[::-1,:]), axis=0)
        n = np.concatenate((n,n[:,::-1]), axis=1)

        n = np.roll(np.roll(n, self.Nx//4, axis = 0), self.Ny//4, axis = 1)

        n = (fft.ifft2(fft.fftshift(self.Sq)*fft.fft2(fft.fftshift(n)))).reshape((-1,))

        return np.real(n)

    def E_tot(self, n):
        """
            Calculate the total energy for a given density configuration n
        """

        ns = self.make_n(n)
        E = np.vdot(ns*(self.dx*self.dy), 0.5*(fft.ifft2(fft.fftshift(self.Gq)*fft.fft2((ns).reshape((self.Nx,self.Ny))))).reshape((-1,)))*(2*self.L)**2
        E += np.dot(ns*(self.dx*self.dy), self.Phi_ext) + np.sum(self.E_int(ns))*(self.dx*self.dy)

        return np.real(E)/(2*self.L)**2*self.energy_scale

    def jac_E_tot(self, n):
        """
            Calcu late the gradient of the energy function for a given configuration n.
        """

        ns = self.make_n(n)

        jac_G = fft.fftshift(self.Gq)*np.conjugate(fft.fft2((ns*(self.dx*self.dy)).reshape((self.Nx,self.Ny))))
        jac_G = fft.ifft2(fft.fftshift(self.Sq)*jac_G)
        jac_G = np.roll(np.roll(jac_G, -1, axis = 0), -1, axis = 1)*(2*self.L)**2

        jac_ext = fft.fft2(((self.Phi_ext)*(self.dx*self.dy)).reshape((self.Nx,self.Ny)))
        jac_ext = fft.ifft2(fft.fftshift(self.Sq)*jac_ext)

        jac = fft.fft2(((self.mu(ns))*(self.dx*self.dy)).reshape((self.Nx,self.Ny)))
        jac = fft.ifft2(fft.fftshift(self.Sq)*jac)
        jac = jac[::-1,::-1]
        jac += jac_G + jac_ext

        jac = 4*jac[self.Nx//4:3*self.Nx//4,self.Ny//4:3*self.Ny//4]
        jac = jac.reshape((-1,))

        return np.ascontiguousarray(np.real(jac))/(2*self.L)**2*self.energy_scale

    def init_n(self, eta = 0.15, uniform = False, int_dot = False):
        """
            Helper function to prepare initial density configurations.
            If uniform, create a uniform \nu=1 state with noise eta
            Otherwise we initialize a noise \nu=0 and \nu=1 state under appropriate gates.
        """

        if uniform:
            n = (1 - eta) * (np.ones((self.Nx//2,self.Ny//2)) + eta * np.random.rand(self.Nx//2,self.Ny//2)).reshape((-1,))
        else:
            x = self.TFdata.X + 1e-8
            y = self.TFdata.Y + 1e-8
            n = 0.5 + 0.5*np.sign(x*y)[self.Nx//4:3*self.Nx//4, self.Ny//4:3*self.Ny//4]
            if int_dot:
                n *= ((x**2 + y**2) > 16)[self.Nx//4:3*self.Nx//4, self.Ny//4:3*self.Ny//4]
                n +=  ((x**2 + y**2) < 9)[self.Nx//4:3*self.Nx//4, self.Ny//4:3*self.Ny//4]
            n = n.reshape((-1,))
            n = (1 - eta) * n + eta * np.random.rand(len(n))

        return n

    def find_opt(self, n0 = [], n_iter = None, opt_factor = 1.):
        """
            Global optimization.
            Runs a (bounded) basinhopping optimizer for n_iter reps.
            Internally the global optimizer uses the L-BFGS-B local optimizer.
        """

        class MyBounds:

            def __init__(self, xmax=[1.1,1.1], xmin=[-1.1,-1.1] ):
                self.xmax = np.array(xmax)
                self.xmin = np.array(xmin)
            def __call__(self, **kwargs):
                x = kwargs["x_new"]
                tmax = bool(np.all(x <= self.xmax))
                tmin = bool(np.all(x >= self.xmin))
                return tmax and tmin

        if n_iter == None:
            n_iter = self.n_iter

        if len(n0) == 0:
            n = self.init_n()
        else:
            n = np.copy(n0)

        my_bounds = MyBounds(xmax=[self.n_upper for _ in n], xmin=[self.n_lower for _ in n])

        my_kwargs = {"method":"L-BFGS-B", "jac":self.jac_E_tot, "bounds":tuple([(self.n_lower,self.n_upper) for _ in n])}

        start_time = time.time()

        self.opt_engine = optimize.basinhopping(self.E_tot, n, minimizer_kwargs = my_kwargs, niter = n_iter, T = self.T, accept_test = my_bounds)

        end_time = time.time()

        n = self.opt_engine['x']

        print(f"opt time = {np.round(end_time - start_time)} s")

        return n

    def find_local_opt(self, n0=[]):
        """
            Local optimization.
            Runs a (bounded) L-BFGS-B optimizer.
        """
        if len(n0) == 0:
            n = self.init_n()
        else:
            n = np.copy(n0)

        my_bounds = tuple([(self.n_lower, self.n_upper) for _ in n])

        start_time = time.time()

        self.opt_engine = optimize.minimize(self.E_tot, n, method='L-BFGS-B', jac=self.jac_E_tot, bounds=my_bounds)

        end_time = time.time()

        n = self.opt_engine['x']

        print(f"opt time = {np.round(end_time - start_time)} s")

        return n


# -*- coding: utf-8 -*-

import setuptools

short_description = "Performing Thomas Fermi calculations of two dimensional electronic systems"
try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except: # noqa
    long_description = short_description


if __name__ == "__main__":
    setuptools.setup(
        name='thomasfermi',
        version=0.7,
        description=short_description,
        author='Taige Wang',
        license='MIT',
        packages=['thomasfermi'],
        include_package_data=True,
        python_requires='>=3.6',
        install_requires=[
            'numpy>=1.7',
            'scipy>=1.7',
        ],
        extras_require={
            'tests': [
                'pytest'
            ],
        },

        tests_require=[
            'pytest'
        ],

        zip_safe=True,
        long_description=long_description,
        long_description_content_type="text/markdown"
    )
